use strict;
use Test::More;
use Test::Exception;

use Net::Bullfinch;

# my $bf = Net::Bullfinch->new(host => 'kestrel.example.com',timeout=>-1);
#     my $rt=$bf->send(
#         request_queue => 'on-demand-sql-0',
#         request => { statement => 'getDeployed' },
#         response_queue_suffix => 'cory',
#     );
ok(1);

done_testing;
